==> bfcc-0.1.tar.gz: bfcc/compiler.c <==
/*
    Copyright (c) 2002 Florian Wesch <fw@dividuum.de>. All Rights Reserved.     
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    
    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
*/    


#include <stdio.h> 
#include "node.h"
#include "bfasm.h"
#include "y.tab.h" 
#include "common.h"

// Kurzschreibweisen
// p ist immer die aktuelle node und muss nicht bergeben werden
#define OP(num)         N_OP(p, (num))
#define OPTYPE(num)     N_OPTYPE(p, (num))
#define OPVALUE(num)    N_OPVALUE(p, (num))
#define CONSTVALUE      N_VALUE(p)

#define IDSYMLOC        GetSymLoc(st, N_VARNAME(p))
#define OPSYMLOC(num)   GetSymLoc(st, N_VARNAME(N_OP(p, (num))))

char *optname(int op) {
    char *name="unknown?";
    switch(op) { 
        case WHILE: name="while";    break;
        case IF:    name="if";       break;
        case INPUT: name="input";    break;
        case PRINT: name="print";    break;
        case '=':   name="=";        break;
        case ';':   name=";";        break;
        case '+':   name="+";        break; 
        case '-':   name="-";        break; 
        case '*':   name="*";        break; 
        case '/':   name="/";        break;
        case '<':   name="<";        break;
        case '>':   name=">";        break;
        case GE:    name=">=";       break;
        case LE:    name="<=";       break;
        case NE:    name="!=";       break;
        case EQ:    name="==";       break;
        case OR:    name="||";       break;
        case AND:   name="&&";       break;
    }
    return name;
}

int optimize(nodeType *p) {
    int o;
    if (!p) 
        return 0; 
    
    switch(p->type) { 
        case typeCon:        
            // Hier gibt es nicht zu optimieren
            break; 
        case typeId:         
            // Hier gibt es nicht zu optimieren
            break; 

        case typeOpr: 
            for (o=0; o < N_NOPS(p); o++)
                optimize(OP(o));

            switch(N_OPER(p)) { 
                case WHILE: 
                    if (OPTYPE(0) == typeCon) {         // While Bedingung eine Konstante?
                        int cond = OPVALUE(0);

                        if (cond) {
                            info("achtung: while bedingung immer wahr\n");
                        } else {
                            info("achtung: while bedingung immer falsch\n");
                            freeNode(OP(0));            // Bedingungsnode lschen
                            freeNode(OP(1));            // while stmt lschen
                            MakeDummyNode(p);           // While Node zur Dummynode machen
                        }
                    }
                    break;
                case IF: 
                    if (OPTYPE(0) == typeCon) {         // If Bedingung eine Konstante?
                        int cond = OPVALUE(0);
                        freeNode(OP(0));                // Bedingungsnode lschen

                        info("optimiere if cond (%d) => ", cond);
                        if (cond) {                     // IF Bedingung wahr?
                            info("Op ist immer wahr!\n");
                            if (p->as.opr.nops > 2) {   // berflssiges else stmt lschen
                                freeNode(OP(2));
                            }
                            ReplaceNode(p, OP(1));      // Ersetzt If Node durch Node des If Teils
                        } else {                        // IF Bedingung falsch?
                            info("Op ist immer falsch!\n");
                            freeNode(OP(1));            // if stmt lschen
                            info("%d\n", p->as.opr.nops);
                            if (p->as.opr.nops > 2) {   // gibt es ein else stmt?
                                ReplaceNode(p, OP(2));  // Ersetzt If Node durch Node des Else Teils
                            } else {
                                MakeDummyNode(p);       // If Node zur Dummynode machen
                            }
                        }
                    }
                    break;

                case INPUT:
                case PRINT: 
                case '=':  
                case ';':
                    // Hier gibt es nicht zu optimieren
                    break;


                case '+': case '-': case '*': case '/':
                case '<': case '>':
                case GE:  case LE:  case NE: case EQ: 
                case OR:  case AND:
                    if (OPTYPE(0) == typeCon && OPTYPE(1) == typeCon) {
                        int v0 = OPVALUE(0);
                        int v1 = OPVALUE(1);
                        info("optimiere %d %s %d => ", v0, optname(N_OPER(p)), v1);
                        freeNode(OP(0));
                        freeNode(OP(1));
                        p->type = typeCon;
                        switch(N_OPER(p)) { 
                            case '+': N_VALUE(p) = v0 +  v1; break;
                            case '-': N_VALUE(p) = v0 -  v1; break;
                            case '*': N_VALUE(p) = v0 *  v1; break;
                            case '/': N_VALUE(p) = v0 /  v1; break;
                            case '<': N_VALUE(p) = v0 <  v1; break;
                            case '>': N_VALUE(p) = v0 >  v1; break;
                            case GE:  N_VALUE(p) = v0 >= v1; break;
                            case LE:  N_VALUE(p) = v0 <= v1; break;
                            case NE:  N_VALUE(p) = v0 != v1; break;
                            case EQ:  N_VALUE(p) = v0 == v1; break;
                            case OR:  N_VALUE(p) = v0 || v1; break;
                            case AND: N_VALUE(p) = v0 && v1; break;
                        }
                        info("%d\n", N_VALUE(p));
                    }
                    break;
                default: 
                    fatalerror("opt defaultcase?");
            } 
    } 
    return 0;
}

int dumpdotty_recursive(nodeType *p) {
    int o;
    if (!p) {
        info("NULL [ node=\"NULL\" ]\n");
        return 0; 
    }
    
    switch(p->type) { 
        case typeCon:        
            info("node%p [ label=\"%p type=CONST value=%d\" ];\n", p, p, N_VALUE(p));
            break; 
        case typeId:         
            info("node%p [ label=\"%p type=ID varname=%s\" ];\n", p, p, p->as.id.varname);
            break; 
        case typeOpr: 
            for (o=0; o < p->as.opr.nops; o++)
                dumpdotty_recursive(OP(o));
            info("node%p [ label=\"%p type=OPR opt=%s optnum=%d\" ];\n", p, p, optname(p->as.opr.oper), p->as.opr.nops);
            for (o=0; o < p->as.opr.nops; o++)
                info("node%p -> node%p [ label=\"%d\" ];\n", p, OP(o), o);
            break;
    } 
    return 0;
}

int dumpdotty(nodeType *p) {
    info("digraph test123 {\n");
    dumpdotty_recursive(p);
    info("}\n");
	return 0;
}
                    
int compile(nodeType *p, BFADDR result) { 
    // 2 mgliche temporre Speicher fr variablen
    BFADDR temp1;
    BFADDR temp2;

    int i;
    
    if (!p) 
        return 0; 
    
    switch(p->type) { 
        case typeCon:        
            BFCALL2(set_const, result, CONSTVALUE);
            break; 
        case typeId:         
            if (result == INVALIDADDR)
                fatalerror("typeID mit INVALIDADDR");
            else 
                BFCALL2(copy, IDSYMLOC, result);
            break; 
        case typeOpr: 
            switch(p->as.opr.oper) { 
                case WHILE: 
                    temp1 = BFNEWVAR;
                    compile(OP(0), temp1); 
                    while_begin(temp1);
                        compile(OP(1), INVALIDADDR); 
                        compile(OP(0), temp1); 
                    while_end(temp1);
                    BFREMOVEVAR(temp1);
                    break; 
                case IF: 
                    temp1 = BFNEWVAR;
                    compile(OP(0), temp1); 
                    if (p->as.opr.nops > 2) {  // if else 
                        ifelse_begin(temp1);
                            compile(OP(1), INVALIDADDR);
                        ifelse_else(temp1);
                            compile(OP(2), INVALIDADDR);
                        ifelse_end;
                    } else { //  if
                        if_begin(temp1)
                            compile(OP(1), INVALIDADDR);
                        if_end;
                    } 
                    BFREMOVEVAR(temp1);
                    break; 
                case INPUT:      
                    BFCALL1(input, result);
                    break; 
                case PRINT:      
                    switch (OPTYPE(0)) {
                        case typeId:
                            BFCALL1(print, OPSYMLOC(0));
                            break;
                        case typeCon:
                            temp1 = BFNEWVAR;
                            BFCALL2(set_const, temp1, OPVALUE(0));
                            BFCALL1(print, temp1);
                            BFREMOVEVAR(temp1);
                            break;
                        case typeOpr:
                            temp1 = BFNEWVAR;
                            compile(OP(0), temp1);
                            BFCALL1(print, temp1);
                            BFREMOVEVAR(temp1);
                            break;                            
                    }
                    break; 
                case ';':
                    for (i=0; i < p->as.opr.nops; i++)
                        compile(OP(i), INVALIDADDR); 
                    break;
                case '=':        
                    if (OPTYPE(1) == typeCon) {
                        BFCALL2(set_const, OPSYMLOC(0), OPVALUE(1));
                    } else {
                        temp1 = BFNEWVAR;
                        compile(OP(1), temp1);
                        BFCALL2(move_destroy, temp1, OPSYMLOC(0));
                        BFREMOVEVAR(temp1);
                    }
                    break; 
                case '+':
                    temp1 = BFNEWVAR;
                    compile(OP(0), result); 
                    compile(OP(1), temp1); 
                    BFCALL2(add_destroy, result, temp1);
                    BFREMOVEVAR(temp1);
                    break;
                case '-':
                    temp1 = BFNEWVAR;
                    compile(OP(0), result); 
                    compile(OP(1), temp1); 
                    BFCALL2(substract_destroy, result, temp1);
                    BFREMOVEVAR(temp1);
                    break;
                case '*':
                    temp1 = BFNEWVAR;
                    compile(OP(0), temp1); 
                    compile(OP(1), result); 
                    BFCALL2(mul, result, temp1);
                    BFREMOVEVAR(temp1);
                    break; 
                case '/':
                    fatalerror("/ not implemented");
                default: 
                    temp1 = BFNEWVAR;
                    temp2 = BFNEWVAR;
                    compile(OP(0), temp1); 
                    compile(OP(1), temp2); 
                    switch(p->as.opr.oper) { 
                        case '<':   BFCALL3(Less_destroy,           temp1, temp2, result); break;
                        case '>':   BFCALL3(Greater_destroy,        temp1, temp2, result); break;
                        case GE:    BFCALL3(GreaterOrEqual_destroy, temp1, temp2, result); break; 
                        case LE:    BFCALL3(LessOrEqual_destroy,    temp1, temp2, result); break; 
                        case NE:    BFCALL3(notEqual_destroy,       temp1, temp2, result); break; 
                        case EQ:    BFCALL3(Equal_destroy,          temp1, temp2, result); break; 
                        case OR:    BFCALL3(or,                     temp1, temp2, result); break; 
                        case AND:   BFCALL3(and,                    temp1, temp2, result); break; 
                        default:    fatalerror("default mit %c", p->as.opr.oper);
                    } 
                    BFREMOVEVAR(temp2);
                    BFREMOVEVAR(temp1);
            } 
    } 
    return 0;
}

==> bfcc-0.2.tar.gz: bfcc/compiler.c <==
/*
    Copyright (c) 2002 Florian Wesch <fw@dividuum.de>. All Rights Reserved.     
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    
    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
*/    


#include <stdio.h> 
#include "node.h"
#include "symtable.h"
#include "bfasm.h"
#include "y.tab.h" 
#include "common.h"

// Kurzschreibweisen
// p ist immer die aktuelle node und muss nicht bergeben werden
#define OP(num)         N_OP(p, (num))
#define OPTYPE(num)     N_OPTYPE(p, (num))
#define OPVALUE(num)    N_OPVALUE(p, (num))

#define CONSTVALUE      N_VALUE(p)
#define INDEX           N_INDEX(p)  

#define IDSYM           GetSym(st, N_VARNAME(p))
#define OPSYM(num)      GetSym(st, N_VARNAME(N_OP(p, (num))))

char *optname(int op) {
    char *name="unknown?";
    switch(op) { 
        case WHILE: name="while";    break;
        case IF:    name="if";       break;
        case INPUT: name="input";    break;
        case PRINT: name="print";    break;
        case '=':   name="=";        break;
        case ';':   name=";";        break;
        case '+':   name="+";        break; 
        case '-':   name="-";        break; 
        case '*':   name="*";        break; 
        case '/':   name="/";        break;
        case '<':   name="<";        break;
        case '>':   name=">";        break;
        case GE:    name=">=";       break;
        case LE:    name="<=";       break;
        case NE:    name="!=";       break;
        case EQ:    name="==";       break;
        case OR:    name="||";       break;
        case AND:   name="&&";       break;
    }
    return name;
}

int optimize(nodeType *p) {
    int o;
    if (!p) 
        return 0; 
    
    switch(N_TYPE(p)) { 
        case typeCon:        
            // Hier gibt es nicht zu optimieren
            break; 
        case typeId:         
            optimize(OP(1)); // index
            break; 

        case typeOpr: 
            for (o=0; o < N_NOPS(p); o++)
                optimize(OP(o));

            switch(N_OPER(p)) { 
                case WHILE: 
                    if (OPTYPE(0) == typeCon) {         // While Bedingung eine Konstante?
                        int cond = OPVALUE(0);

                        if (cond) {
                            info("achtung: while bedingung immer wahr\n");
                        } else {
                            info("achtung: while bedingung immer falsch\n");
                            freeNode(OP(0));            // Bedingungsnode lschen
                            freeNode(OP(1));            // while stmt lschen
                            MakeDummyNode(p);           // While Node zur Dummynode machen
                        }
                    }
                    break;
                case IF: 
                    if (OPTYPE(0) == typeCon) {         // If Bedingung eine Konstante?
                        int cond = OPVALUE(0);
                        freeNode(OP(0));                // Bedingungsnode lschen

                        info("optimiere if cond (%d) => ", cond);
                        if (cond) {                     // IF Bedingung wahr?
                            info("Op ist immer wahr!\n");
                            if (N_NOPS(p) > 2) {        // berflssiges else stmt lschen
                                freeNode(OP(2));
                            }
                            ReplaceNode(p, OP(1));      // Ersetzt If Node durch Node des If Teils
                        } else {                        // IF Bedingung falsch?
                            info("Op ist immer falsch!\n");
                            freeNode(OP(1));            // if stmt lschen
                            info("%d\n", N_NOPS(p));
                            if (N_NOPS(p) > 2) {   // gibt es ein else stmt?
                                ReplaceNode(p, OP(2));  // Ersetzt If Node durch Node des Else Teils
                            } else {
                                MakeDummyNode(p);       // If Node zur Dummynode machen
                            }
                        }
                    }
                    break;

                case DIM:
                case INPUT:
                case PRINT: 
                case '=':  
                case ';':
                    // Hier gibt es nicht zu optimieren
                    break;


                case '+': case '-': case '*': case '/':
                case '<': case '>':
                case GE:  case LE:  case NE: case EQ: 
                case OR:  case AND:
                    if (OPTYPE(0) == typeCon && OPTYPE(1) == typeCon) {
                        int v0 = OPVALUE(0);
                        int v1 = OPVALUE(1);
                        info("optimiere %d %s %d => ", v0, optname(N_OPER(p)), v1);
                        freeNode(OP(0));
                        freeNode(OP(1));
                        N_TYPE(p) = typeCon;
                        switch(N_OPER(p)) { 
                            case '+': N_VALUE(p) = v0 +  v1; break;
                            case '-': N_VALUE(p) = v0 -  v1; break;
                            case '*': N_VALUE(p) = v0 *  v1; break;
                            case '/': N_VALUE(p) = v0 /  v1; break;
                            case '<': N_VALUE(p) = v0 <  v1; break;
                            case '>': N_VALUE(p) = v0 >  v1; break;
                            case GE:  N_VALUE(p) = v0 >= v1; break;
                            case LE:  N_VALUE(p) = v0 <= v1; break;
                            case NE:  N_VALUE(p) = v0 != v1; break;
                            case EQ:  N_VALUE(p) = v0 == v1; break;
                            case OR:  N_VALUE(p) = v0 || v1; break;
                            case AND: N_VALUE(p) = v0 && v1; break;
                        }
                        info("%d\n", N_VALUE(p));
                    }
                    break;
                default: 
                    fatalerror("opt defaultcase?");
            } 
    } 
    return 0;
}

void FetchSymbols(nodeType *p) {
    int o;
    Symbol *s;

    if (!p)
        return;

    switch(p->type) { 
        case typeCon:        
            break; 
        case typeId:         
            FetchSymbols(INDEX);
            s = GetSymbol(st, N_VARNAME(p));
            if (N_TYPE(INDEX) == typeCon) {
                if (!s) {
                    NewSymbol(st, N_VARNAME(p), N_VALUE(INDEX)+1);
                } else {
                    if (N_VALUE(INDEX) > 0 && 
                        SymbolGrowSize(st, N_VARNAME(p), N_VALUE(INDEX)+1)) {
                        info("Warning: Implicit dim'ing of %s: new size %d\n", N_VARNAME(p), N_VALUE(INDEX));
                    }
                }
            } else if (!s) {
                fatalerror("%s used with index expression, but storage size of %s unknown. you must dim %s first.",
                        N_VARNAME(p),N_VARNAME(p),N_VARNAME(p));
            }
            break; 
        case typeOpr: 
            if (N_OPER(p) == DIM) {
                if (OPTYPE(1) != typeCon) // dim must have a constant value
                    fatalerror("dim %s with nonconstant value", N_VARNAME(OP(0)));

                s = GetSymbol(st, N_VARNAME(OP(0)));

                if (s)
                    fatalerror("%s dim'ed twice?", N_VARNAME(OP(0)));

                NewSymbol(st, N_VARNAME(OP(0)), N_VALUE(OP(1)));
            } else {
                for (o=0; o < N_NOPS(p); o++)
                    FetchSymbols(OP(o));
            }
            break;

    }
}

int dumpdotty_recursive(nodeType *p) {
    int o;
    if (!p) {
        info("NULL [ node=\"NULL\" ]\n");
        return 0; 
    }
    
    switch(N_TYPE(p)) { 
        case typeCon:        
            info("node%p [ label=\"%p type=CONST value=%d\" ];\n", p, p, N_VALUE(p));
            break; 
        case typeId:         
            info("node%p [ label=\"%p type=ID varname=%s\" ];\n", p, p, p->as.id.varname);
            break; 
        case typeOpr: 
            for (o=0; o < N_NOPS(p); o++)
                dumpdotty_recursive(OP(o));
            info("node%p [ label=\"%p type=OPR opt=%s optnum=%d\" ];\n", p, p, optname(p->as.opr.oper), p->as.opr.nops);
            for (o=0; o < N_NOPS(p); o++)
                info("node%p -> node%p [ label=\"%d\" ];\n", p, OP(o), o);
            break;
    } 
    return 0;
}

int dumpdotty(nodeType *p) {
    info("digraph test123 {\n");
    dumpdotty_recursive(p);
    info("}\n");
	return 0;
}
                    
int compile(nodeType *p, BFADDR result) { 
    // 2 mgliche temporre Speicher fr variablen
    BFADDR temp1;
    BFADDR temp2;
    Symbol *sym;

    int i;
    
    if (!p) 
        return 0; 
    
    switch(N_TYPE(p)) { 
        case typeCon:        
            BFCALL2(set_const, result, CONSTVALUE);
            break; 
        case typeId:         
            if (result == INVALIDADDR)
                fatalerror("typeID mit INVALIDADDR");
            else {
                sym = IDSYM;
                if (SymIsArray(sym)) {
                    temp1 = BFNEWVAR;
                    compile(INDEX, temp1);
                    BFCALL3(ArrayGet, sym->mempos, temp1, result);
                    BFREMOVEVAR(temp1);
                } else {
                    BFCALL2(copy, sym->mempos, result);
                }
            }
            break; 
        case typeOpr: 
            switch(N_OPER(p)) { 
                case WHILE: 
                    temp1 = BFNEWVAR;
                    compile(OP(0), temp1); 
                    while_begin(temp1);
                        compile(OP(1), INVALIDADDR); 
                        compile(OP(0), temp1); 
                    while_end(temp1);
                    BFREMOVEVAR(temp1);
                    break; 
                case IF: 
                    temp1 = BFNEWVAR;
                    compile(OP(0), temp1); 
                    if (N_NOPS(p) > 2) {  // if else 
                        ifelse_begin(temp1);
                            compile(OP(1), INVALIDADDR);
                        ifelse_else(temp1);
                            compile(OP(2), INVALIDADDR);
                        ifelse_end;
                    } else { //  if
                        if_begin(temp1)
                            compile(OP(1), INVALIDADDR);
                        if_end;
                    } 
                    BFREMOVEVAR(temp1);
                    break; 
                case INPUT:      
                    BFCALL1(input, result);
                    break; 
                case PRINT:      
                    switch (OPTYPE(0)) {
                        case typeCon:
                            temp1 = BFNEWVAR;
                            BFCALL2(set_const, temp1, OPVALUE(0));
                            BFCALL1(print, temp1);
                            BFREMOVEVAR(temp1);
                            break;
                        case typeId:
                        case typeOpr:
                            temp1 = BFNEWVAR;
                            compile(OP(0), temp1);
                            BFCALL1(print, temp1);
                            BFREMOVEVAR(temp1);
                            break;                            
                    }
                    break; 
                case ';':
                    for (i=0; i < N_NOPS(p); i++)
                        compile(OP(i), INVALIDADDR); 
                    break;
                case '=':        
                    sym = OPSYM(0);
                    if (SymIsArray(sym)) {
                        temp1 = BFNEWVAR;
                        temp2 = BFNEWVAR;
                        compile(N_INDEX(OP(0)), temp1);
                        compile(OP(1), temp2);
                        BFCALL3(ArraySet, sym->mempos, temp1, temp2);
                        BFREMOVEVAR(temp2);
                        BFREMOVEVAR(temp1);
                    } else {
                        if (OPTYPE(1) == typeCon) {
                            BFCALL2(set_const, sym->mempos, OPVALUE(1));
                        } else {
                            temp1 = BFNEWVAR;
                            compile(OP(1), temp1);
                            BFCALL2(move_destroy, temp1, sym->mempos);
                            BFREMOVEVAR(temp1);
                        }
                    }
                    break; 
                case '+':
                    temp1 = BFNEWVAR;
                    compile(OP(0), result); 
                    compile(OP(1), temp1); 
                    BFCALL2(add_destroy, result, temp1);
                    BFREMOVEVAR(temp1);
                    break;
                case '-':
                    temp1 = BFNEWVAR;
                    compile(OP(0), result); 
                    compile(OP(1), temp1); 
                    BFCALL2(substract_destroy, result, temp1);
                    BFREMOVEVAR(temp1);
                    break;
                case '*':
                    temp1 = BFNEWVAR;
                    compile(OP(0), temp1); 
                    compile(OP(1), result); 
                    BFCALL2(mul, result, temp1);
                    BFREMOVEVAR(temp1);
                    break; 
                case '/':
                    fatalerror("/ not implemented");
                case DIM:
                    sym = OPSYM(0);
                    if (!SymIsArray(sym))
                        fatalerror("dim without array");
                    BFCALL1(ArrayInit, sym->mempos);
                    break;
                default: 
                    temp1 = BFNEWVAR;
                    temp2 = BFNEWVAR;
                    compile(OP(0), temp1); 
                    compile(OP(1), temp2); 
                    switch(N_OPER(p)) { 
                        case '<':   BFCALL3(Less_destroy,           temp1, temp2, result); break;
                        case '>':   BFCALL3(Greater_destroy,        temp1, temp2, result); break;
                        case GE:    BFCALL3(GreaterOrEqual_destroy, temp1, temp2, result); break; 
                        case LE:    BFCALL3(LessOrEqual_destroy,    temp1, temp2, result); break; 
                        case NE:    BFCALL3(notEqual_destroy,       temp1, temp2, result); break; 
                        case EQ:    BFCALL3(Equal_destroy,          temp1, temp2, result); break; 
                        case OR:    BFCALL3(or,                     temp1, temp2, result); break; 
                        case AND:   BFCALL3(and,                    temp1, temp2, result); break; 
                        default:    fatalerror("default mit %c", p->as.opr.oper);
                    } 
                    BFREMOVEVAR(temp2);
                    BFREMOVEVAR(temp1);
            } 
    } 
    return 0;
}
